#include <SDL/SDL.h>
#include "title_screen.h"
#include "gameover.h"
#include "font/3x5.h"
#include "font/8x12.h"
#include "blocks.h"
#include <portaudio.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PREVIEW_WIDTH 128
#define PREVIEW_HEIGHT (PREVIEW_WIDTH * 1.5625)

#define BLOCK_WIDTH ((PREVIEW_WIDTH) / 32)
#define BLOCK_HEIGHT ((PREVIEW_HEIGHT) / 50)

#undef main
//R G O Y G O R
enum COLOR {BLACK, RED, GREEN, ORANGE, YELLOW};

unsigned char image[1600];
const unsigned char palette[5][3] = {{0, 0, 0}, {255, 0, 0}, {0, 255, 0}, {255, 145, 0}, {255, 255, 0}};
short samples[6400];
int fbp = 0;

unsigned char board[7][16];

int px = 2, py = 0;
int type = 0;
int rotation = 0;
int score = 0;

void putpixel(SDL_Surface *s, int x, int y, int r, int g, int b){
	char *pixels = s->pixels;
	pixels[y * s->pitch + x * 3] = b;
	pixels[y * s->pitch + x * 3 + 1] = g;
	pixels[y * s->pitch + x * 3 + 2] = r;
}

void putblock(SDL_Surface *s, int x, int y, int r, int g, int b){
	for(int xp = 0;xp < BLOCK_WIDTH;xp++){
		for(int yp = 0;yp < BLOCK_HEIGHT;yp++){
			putpixel(s, x * BLOCK_WIDTH + xp, y * BLOCK_HEIGHT + yp, r, g, b);
		}
	}
}

void fputpixel(int x, int y, int c){
	if(x < 0 || x > 31 || y < 0 || y > 47) return;
	if(c < 0) c = 0;
	if(c > 5) c = 5;
	image[x * 50 + y] = c;
}

void fclear(){
	memset(image, 0, 1600);
}

void draw_next(int x, int y, int next){
	for(int xx = 0;xx < 8;xx++){
		for(int yy = 0;yy < 8;yy++){
			fputpixel(xx + x, yy + y, types[next][yy/2 * 4 + xx / 2]);
		}
	}
}

void draw_number(int xoff, int yoff, int font, int color, int number){
	if(number >= 48) number -= 48;
	if(number < 0 || number > 9) return;
	//printf("Drawing number %d\n", number);
	if(font == 0){
		unsigned const char *source = font_3x5[number];
		for(int x = 0;x < 4;x++){
			for(int y = 0;y < 5;y++){
				fputpixel(x + xoff, y + yoff, source[y * 4 + x] * color);
			}
		}
	} else if(font == 1){
		unsigned const char *source = font_8x12[number];
		for(int x = 0;x < 8;x++){
			for(int y = 0;y < 12;y++){
				fputpixel(x + xoff, y + yoff, source[y * 8 + x] * color);
			}
		}
	}
}

void draw_score(int x, int y, int font, int score){
	if(score > 99) score = 99;
	if(score < 0) score = 0;
	//printf("Drawing score %d\n", score);
	char buf[4];
	sprintf(buf, "%d", score);
	fflush(stdout);
	for(int i = 0;i < strlen(buf);i++){
		draw_number(x + i * (font? 8 : 4), y, font, GREEN, buf[i]);
	}
}

void render_board(SDL_Surface *s, int next){
	for(int x = 0;x < 7;x++){
		for(int y = 0;y < 16;y++){
			for(int xx = 0;xx < 3;xx++){
				for(int yy = 0;yy < 3;yy++){
					if(11 + xx + x * 3 == 31 && yy + y * 3 % 2 == 0) continue;
					fputpixel(11 + xx + x * 3, yy + y * 3, board[x][y]);
				}
			}
		}
	}
	draw_next(0, 30, next);
}

void render(SDL_Surface *s){
	for(int i = 0;i < 3200;i++){
		if(i % 100 >= 0 && i % 100 < 6){
			if(i > 6){
				samples[i * 2] = -14000;
				samples[i * 2 + 1] = -14000;
			} else {
				samples[i * 2] = 0;
				samples[i * 2 + 1] = 0;
			}
		} else {
			int color = image[1599 - i / 2];
			unsigned const char *rgb = palette[color];
			samples[i * 2] = rgb[0] * 128;
			samples[i * 2 + 1] = rgb[1] * 96;
		}
	}
	for(int x = 0;x < 32;x++){
		for(int y = 0;y < 50;y++){
			int color = image[x * 50 + y];
			unsigned const char *rgb = palette[color];
			putblock(s, x, y, rgb[0], rgb[1], rgb[2]);
		}
	}
	SDL_Flip(s);
}

void put_piece(){
	unsigned char *source = types[type] + (rotation * 16);
	for(int x = 0;x < 4;x++){
		for(int y = 0;y < 4;y++){
			if(x < 0 || x > 6 || y < 0 || y > 15) continue;
			board[x + px][y + py] ^= source[y * 4 + x];
		}
	}
}

int move_block(int dx, int dy){
	unsigned char *source = types[type] + (rotation * 16);
	put_piece();
	for(int x = 0;x < 4;x++){
		for(int y = 0;y < 4;y++){
			int cx = px + dx + x;
			int cy = py + dy + y;
			if(source[y * 4 + x]){
				if(cx > 6 || cx < 0 || cy > 15 || cy < 0){
					printf("Bounds check\n");
					put_piece();
					return 0;
				}
				if(board[cx][cy]){
					printf("Collision\n");
					put_piece();
					return 0;
				}
			}
		}
	}
	px += dx;
	py += dy;
	put_piece();
	return 1;
}

int rotate_block(){
	unsigned char *source = types[type] + ((1 + rotation) % 4 * 16);
	put_piece();
	for(int x = 0;x < 4;x++){
		for(int y = 0;y < 4;y++){
			int cx = px + x;
			int cy = py + y;
			if(cx > 6){
				printf("Wall bump\n");
				put_piece();
				if(!move_block(-1, 0)){
					printf("Failed\n");
					put_piece();
					return 0;
				} else return rotate_block();
			}
			if(cx < 0){
				printf("Wall bump\n");
				put_piece();
				if(!move_block(1, 0)){
					printf("Failed\n");
					put_piece();
					return 0;
				} else return rotate_block();
			}
			if(source[y * 4 + x]){
				if(cy > 15 || cy < 0){
					printf("Bounds check\n");
					put_piece();
					return 0;
				}
				if(board[cx][cy]){
					printf("Collision\n");
					put_piece();
					return 0;
				}
			}
		}
	}
	rotation = (rotation + 1) % 4;
	put_piece();
	return 1;
}

int callback(const void *input, void *output, unsigned long frameCount, const PaStreamCallbackTimeInfo *timeinfo, const PaStreamCallbackFlags statusFlags, void *userData){
	int i = 0;
	//printf("Framecount = %d\n", frameCount);
	while(i < frameCount){
		for(;i < frameCount && fbp < 3200;i++){
			((int32_t*)output)[i] = ((int32_t*)samples)[fbp++];
		}
		if(fbp >= 3199){
			fbp = 0;
		}
	}
	return paContinue;
}

int clear_blocks(){
	int result = 0;
	for(int y = 15;y >= 0;y--){
		int line_count = 0;
		for(int x = 0;x < 7;x++){
			if(board[x][y]){
				line_count++;
			}
		}
		if(line_count == 7){
			result++;
			for(int j = y;j >= 0;j--){
				for(int x = 0;x < 7;x++){
					if(j == 0){
						board[x][j] = 0;
					} else {
						board[x][j] = board[x][j - 1];
					}
				}
			}
		}
	}
	return result;
}

void draw_image(char *img){
	for(int x = 0;x < 32;x++){
		for(int y = 0;y < 48;y++){
			fputpixel(x, y, img[y * 32 + x]);
		}
	}
}

int main(int argc, char **argv){
	if(Pa_Initialize() != paNoError){
		printf("Failed to init portaudio\n");
		return 1;
	}
	PaStream *stream;
	if(Pa_OpenDefaultStream(&stream, 0, 2, paInt16, 48000, 0, callback, NULL) != paNoError){
		printf("Failed to open output audio stream\n");
		return 1;
	}
	SDL_Init(SDL_INIT_EVERYTHING | SDL_INIT_NOPARACHUTE);
	SDL_Surface *screen = SDL_SetVideoMode(PREVIEW_WIDTH, PREVIEW_HEIGHT, 24, 0);
	if(screen == NULL){
		printf("Failed!\n");
		return 1;
	}
	int run = 1;
	draw_image(title_screen);
	render(screen);
	if(Pa_StartStream(stream) != paNoError){
		printf("Error starting stream\n");
		return 1;
	}
	srand(time(NULL));
	SDL_Event ev;
	while(run){
		if(SDL_PollEvent(&ev)){
			if(ev.type == SDL_QUIT){
				run = 0;
			}
			if(ev.type == SDL_KEYDOWN){
				break;
			}
		}
		SDL_Delay(100);
	}
	fclear();
	type = rand() % 7;
	int next = rand() % 7;
	for(int y = 0;y < 48;y++){
		fputpixel(10, y, YELLOW);
		fputpixel(31, y, RED);
	}
	int pause = 0;
	put_piece();
	render_board(screen, 0);
	int soft_down = 0;
	while(run){
		int next_delay = 400 - score * 10;
		while(SDL_PollEvent(&ev)){
			if(ev.type == SDL_QUIT){
				run = 0;
			}
			if(ev.type == SDL_KEYDOWN){
				int key = ev.key.keysym.sym;
				if(key == SDLK_RIGHT){
					move_block(1, 0);
				}
				if(key == SDLK_LEFT){
					move_block(-1, 0);
				}
				if(key == SDLK_UP){
					rotate_block();
				}
				if(key == SDLK_DOWN){
					soft_down = 1;
				}
				if(key == SDLK_SPACE){
					pause = !pause;
				}
			}
			if(ev.type == SDL_KEYUP){
				int key = ev.key.keysym.sym;
				if(key == SDLK_DOWN){
					soft_down = 0;
				}
			}
		}
		if(pause){
			SDL_Delay(100);
			continue;
		}
		if(soft_down) next_delay = 100 - score * 2;
		score += clear_blocks();
		draw_score(score < 10? 5: 1, 16, 0, score);
		if(!move_block(0, 1)){
			if(py == 0){
				SDL_Delay(1000);
				draw_image(gameover);
				draw_score(score < 10 ? 12 : 8, 32, 1, score);
				render(screen);
				SDL_Delay(1000);
				while(run){
					if(SDL_PollEvent(&ev)){
						if(ev.type == SDL_QUIT || ev.type == SDL_KEYDOWN){
							run = 0;
						}
					}
					SDL_Delay(100);
				}
			}
			py = 0;
			type = next;
			next = rand() % 7;
			px = type? 2 : 1;
			put_piece();
		}
		render_board(screen, next);
		render(screen);
		if(next_delay < 50){
			next_delay = soft_down? 40 : 50;
		}
		SDL_Delay(next_delay);
	}
	SDL_Quit();
}